/* leave.c
 *
 * a small utility to remind you when you have to leave
 *
 * warning messages will be printed 30, 15, 10, 5, 1 minutes before you
 * have to leave, and each minute after you wanted to leave up to five
 * additional reminders are printed to tell you that it's really time
 *
 * Copyright (C) 2005 Manuel Tobias Schiller <mala@hinterbergen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * For the full text of the GNU General Public License, please see the
 * file COPYING which should be present in the same directory as this
 * source file.
 */

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <time.h>

#include "config.h"

/* check if we can trust the system's realloc implementation and provide
 * a safe replacement if not */
#ifndef HAVE_REALLOC
void *myrealloc(void *ptr, size_t size)
{
	if (ptr == NULL) {
		/* check for the malloc-case */
		if (size)
			return malloc(size);
		/* ok, can't do much here */
		return NULL;
	}
	/* check for the free-case */
	if (!size) {
		free(ptr);
		return NULL;
	}
	/* we trust that realloc does its job in the non-degenerate case */
	return realloc(ptr, size);
}
#define realloc myrealloc
#endif

static void sigalarm(int signum)
{
	/* re-establish a handler for SIGALRM (might be unneccessary on
	 * some systems) */
	signal(SIGALRM, sigalarm);
	return;
}

/* make the alarm go off at the right times and print a message to stderr
 * 
 * note that we check the return value of fprintf and fflush carefully to
 * find out if the standard error stream is still available - if it's not,
 * the session might have been terminated early and we're just hanging
 * there, doing no-one any good, eating up resources */
static int doalarm(time_t when)
{
	time_t now = 0;
	time_t remaining;

	/* close stdout - we won't need it anymore */
	fclose(stdout);

	/* establish a handler for SIGALRM */
	signal(SIGALRM, sigalarm);
	now = time(NULL);
	while (now < when) {
		remaining = when - now;
		/* set an alarm to go off 30, 15, 10, 5, 1, 0 minutes
		 * before it's time to leave */
		if ((when - 30 * 60) > now) alarm(remaining - 30 * 60);
		else if ((when - 15 * 60) > now) alarm(remaining - 15 * 60);
		else if ((when - 10 * 60) > now) alarm(remaining - 10 * 60);
		else if ((when -  5 * 60) > now) alarm(remaining -  5 * 60);
		else if ((when -  1 * 60) > now) alarm(remaining -  1 * 60);
		else if ((when -  0 * 60) > now) alarm(remaining -  0 * 60);
		/* sleep until we get an alarm */
		pause();
		/* get the new time */
		now = time(NULL);
		remaining = when - now;
		/* how many minutes are left? */
		if (remaining % 60 >= 30) remaining += 60;
		remaining /= 60;
		/* print a message to stderr */
		if (fprintf(stderr, "\aIt's time to leave ") < 0)
			exit(1);
		/* we handle underflows in remaining by making sure that
		 * less than a day remains - otherwise we assume that an
		 * underflow occurred */
		if ((remaining < 24 * 60) && (remaining > 1)) {
			if (fprintf(stderr, "in %d minutes.\n",
					(int) remaining) < 0)
				exit(1);
		} else if (remaining == 1) {
			if (fprintf(stderr, "in %d minute.\n",
					(int) remaining) < 0)
				exit(1);
		} else {
			if (fprintf(stderr, "now.\n") < 0)
				exit(1);
		}
		if (fflush(stderr) != 0)
			exit(1);
	}
	/* we will tell the user to leave four more times */
	for (now = 3; now--; ) {
		alarm(60);
		pause();
		if (fprintf(stderr, "\aIt's time to leave now.\n") < 0)
			exit(1);
		if (fflush(stderr) != 0)
			exit(1);
	}
	alarm(60);
	pause();
	if (fprintf(stderr, "\aIt's time to leave now. " \
	        "This is the last time I'll tell you.\n") < 0)
		exit(1);
	if (fflush(stderr) != 0)
		exit(1);

	exit(0);
}

/* this is a complete geline replacement for systems that don't have it
 * the implementation is dumb and slow, but it should work */
static int mygetline(char **lineptr, size_t *n, FILE *f)
{
	/* our internal buffer is BUFSZ bytes big, and reallocation will also be
	 *  * to line buffer sizes which are multiples of BUFSZ bytes */
#define BUFSZ   256

	char buf[BUFSZ], *p, *q;
	int c;
	size_t m, total = 0;

	/* make sure our input is valid */
	if ((lineptr == NULL) || (n == NULL)|| (f == NULL))
		return -1;
	/* return on error and end of file */
	if (ferror(f)) return -1;
	if (feof(f)) return -1;

	do {
		/* read in a piece of our line */
		m = 0;
		while ((m < BUFSZ) && !ferror(f) && !feof(f)) {
			buf[m] = c = fgetc(f);
			if (c == EOF) break;
			if (buf[m++] == '\n')
				break;
		}
		/* in case of a "hard" error, we just bail out */
		if (ferror(f)) return -1;

		total += m;

		/* make sure destination is large enough */
		if (total >= *n) {
			size_t newsz = total + 1;
			/* round up to the nearest multiple of BUFSZ
			* to avoid frequent reallocation as longer
			* and longer lines are read in
			* we also make sure we have space for the
			* terminating zero character */
			newsz += (BUFSZ - (newsz % BUFSZ));
			p = (char *) realloc(*lineptr, newsz);
			if ((p == NULL))
				return -1;
			*lineptr = p;
			*n = newsz;
		}

		/* copy to destination */
		q = &(*lineptr)[total - m];
		p = buf;
		while (p < &buf[m]) *q++ = *p++;

		/* check for feof again */
		if (feof(f)) {
			/* if we couldn't read a thing, return -1 */
			if (total == 0) return -1;
			/* we could read something, null-terminate it
			* and return... */
			*q++ = 0;
			return total;
		}
		/* if we just copied our end-of-line character, we're
		* done - note: we did read at least one byte if we reached
		* this point, so *--p is perfectly legal to deference */
	} while (*--p != '\n');

	/* zero-terminate our string - we have space in the buffer! */
	*q++ = 0;
	return total;

#undef BUFSZ
}

/* print a short help message to stdout */
static void help()
{
	printf("usage: leave [[+]hhmm]\n" \
	       "\thhmm\ttime when you have to leave, interpreted on a " \
	       "12 hour clock\n" \
	       "\t+hhmm\thours and minutes from now until you have to " \
	       "leave\n\n" \
	       "\tIf leave is given no argument it asks when you have to " \
	       "leave.\n");

	exit(0);
}

/* identify this version of leave */
static void version()
{
	printf("leave version " PACKAGE_VERSION " ($Id: leave.c 96 2005-12-20 18:32:43Z mala $)\n");
	exit(0);
}

/* try to extract a time from the time spec given by the user */
static time_t parsetime(char *str)
{
	int flag = 0;
	/* we start by ignoring any whitespace we might find */
	while (*str) {
		if (isspace(*str)) str++;
		else break;
	}
	/* make sure there is more input to consume, if there is none,
	 * we assume that the user does not want us to tell him when to
	 * leave - this is useful when leave is run from a login script */
	if (*str == 0) exit(0);
	/* check if we have a plus or minus sign */
	if (*str == '+') flag++, str++;
	else if (*str == '-') {
		/* leave was called in "Douglas Adams"-mode */
		fprintf(stderr, "leave: error 42: Sorry, reverse temporal " \
		        "engineering is not yet implemented.\n");
		exit(42);
	}
	if (strlen(str) < 4) return (time_t) -1;
	if (isdigit(str[0]) && isdigit(str[1]) && \
	        isdigit(str[2]) && isdigit(str[3])) {
		/* ok, we have four digits left - deal with them */
		int h, m;
		struct tm *timep;
		time_t now, when;
		h = 10 * (str[0] - '0') + (str[1] - '0');
		m = 10 * (str[2] - '0') + (str[3] - '0');
		if (h >= 24) return (time_t) -1;
		if (m >= 60) return (time_t) -1;
		h %= 12;
		when = 60 * (m + 60 * h);
		/* if flag is set, we'll be done quickly */
		if (flag) return time(NULL) + when;
		/* ok, we were given absolute time, break down current time */
		now = time(NULL);
		timep = localtime(&now);
		/* make sure our result is within the next twelve hours */
		now = timep->tm_sec+60*(timep->tm_min+60*timep->tm_hour);
		while (now >= when) when += 12 * 3600;
		/* fill in when into timep */
		timep->tm_sec = when % 60;
		when /= 60;
		timep->tm_min = when % 60;
		timep->tm_hour = when / 60;
		/* convert the result back to time_t and return */
		return mktime(timep);
	}
	/* the four characters were not only digits - this is an error */
	return (time_t) -1;
}

/* the main program puts the bits and pieces together... */
int main(int argc, char **argv)
{
	pid_t pid;
	time_t when;
	char *p;
	int i;

	/* search for -h, --help, -v and --version options */
	for (i = argc; i--; ) {
		if ((strcmp(argv[i], "-h") == 0) || \
		        (strcmp(argv[i], "--help") == 0))
			help();
		if ((strcmp(argv[i], "-v") == 0) || \
		        (strcmp(argv[i], "--version") == 0))
			version();
	}

	/* if there are arguments, the first one is our time string,
	 * we'll have to ask for one otherwise */
	if (argc > 1) p = argv[1];
	else {
		size_t n = 0;
		p = 0;
		printf("When do you have to leave? " \
		       "(Answer with '[+]hhmm.')\n");
		if (mygetline(&p, &n, stdin) <= 0) goto inp;
		if (p == NULL) goto inp;

	}
	if ((when = parsetime(p)) == (time_t) -1) goto parse;

	/* close stdin - we don't need it anymore */
	fclose(stdin);

	/* make the alarm stuff happen in the background */
	if ((pid = fork()) == 0) doalarm(when);
	if ((int) pid == -1) perror("leave: fork() failed");

	/* tell the user that and when the alarm is set */
	p = ctime(&when);
	/* remove the trailing newline that ctime puts there */
	p[strlen(p) - 1] = 0;
	printf("Alarm set for %s. (pid %ld)\n", p, (long int) pid);

	return 0;

inp:	/* we couldn't read from stdin... */
	fprintf(stderr, "leave: Couldn't read input.\n");
	return 1;

parse:	/* we failed to parse the input */
	fprintf(stderr, "leave: Couldn't parse time spec.\n");
	return 1;
}
